/* symtable.h
 *
 * symbol table for Ada/CS compiler.
 * Magnus Christensson 1996.
 */
/* modified for project Barracuda, 1997. */

#include <stddef.h>
#include <qstring.h>

/* From Ada/CS compiler defs.h */

/* --- only one of these two should be defined */ 
#define SINGLE_SCOPE_SYMBOL_TABLE 1
#undef MULTIPLE_SCOPE_SYMBOL_TABLE

typedef char *string_t;

#define IDENTIFIER 1

typedef struct string_entry_
{
  string_t ptr;
  unsigned int length;
  unsigned int hash_value;
  int ret_value;		/* IDENTIFIER, RW_ABS,... */
  struct string_entry_* next;
} string_entry_t;

typedef struct id_entry_
{
  struct string_entry_ *name;	/* pointer to string table */
#ifdef MULTIPLE_SCOPE_SYMBOL_TABLE
  int level;			/* scope level */
#endif
  struct attributes_ *attr;	/* attribute info */
  struct id_entry_ *next_symbol; /* used to link all symbols together */
  struct id_entry_ *next_chain;	/* link in hash table chain */
  struct id_entry_ *next_same_name; /* link to next entry with same symbol */
  /* coordinate_t src; */       /* source file coordinates */
} id_entry_t;

typedef struct symbol_table_
{
  size_t size;			/* hash table size minus 1 */
  size_t n_entries;		/* number of entries in the table */
  float threshold;		/* THRESHOLD > 0.0 specifies how full
				   the hash table can get before it
				   must grow. */
  id_entry_t **buckets;		/* pointer to vector of size SIZE + 1 */
  struct symbol_table_ *previous; /* pointer to previous table */
  id_entry_t *all;		/* used to chain all symbols together */
#ifndef MULTIPLE_SCOPE_SYMBOL_TABLE
  int level;			/* scope level corresponding to this table */
#endif
} symbol_table_t;

typedef struct attributes_
{
  QString str;
  bool representation;
  bool catenated;
  bool referenced;
} attributes_t;

/* End from defs.h */

extern char* string (char* ptr, int n);

/* Open a new scope */
extern symbol_table_t* open_scope (void);

/* Close the current scope */
extern void close_scope (void);

/* Create a new (empty) symbol table */
extern symbol_table_t* create (size_t size, float threshold);

/* Remove all entries in table and destroy it */
extern void destroy (symbol_table_t* table);

/* Enter entry in table.
 * Returns: the chain corresponding to entry
 * entry must be an already allocated object with the name structure filled in
 * If enter return present==TRUE then the symbol table is also set up for
 * find_next.
 */
extern id_entry_t* enter (symbol_table_t* table, id_entry_t* entry,
			  bool* present);

/* Returns: the first record with symbol name, or NULL if not present.
 * the lexical level of this symbol is returned through lexical_level */
extern id_entry_t* find_first (string_entry_t* name,
			       int* lexical_level);

/* Returns: the next record with symbol name, or NULL if not present.
 * A call to find_first must precede a call to find_next or the return value
 * is undefined.
 * the lexical level of this symbol is returned through lexical_level */
extern id_entry_t* find_next (int* lexical_level);

/* Associate the attrs records with entry, entry and attrs must be allocated
 * by caller */
extern void set_attributes (id_entry_t* entry, attributes_t* attrs);

/* Get the attributes record associated with entry */
extern attributes_t* get_attributes (const id_entry_t* entry);

/* Push a symbol_table onto the stack */
extern void sts_push (symbol_table_t* table);

/* Pop a symbol_table of the stack */
extern symbol_table_t* sts_pop (void);

/* Get the top of stack */
extern symbol_table_t* sts_current_scope (void);

extern symbol_table_t* current_scope;

/* Initializes the string space */
extern void initialize_string_space (void);

/* Returns: pointer to the entry.
 * Enters name in string table (and allocates space) */
extern string_entry_t* enter_identifier (const char* name, int length);

/* Returns: the entry corresponding to name, or NULL if not present */
extern id_entry_t* find (const symbol_table_t* table,
			 const string_entry_t* name);
