/* MultiLabel.cc
 *
 * Implements the multilabel class.
 * This widget shows a number of labels on a bar with separators
 * between them. The separators can be moved by dragging them with 
 * the mouse.
 *
 * Barracuda 1.0.
 * Copyright (C) 1997 Magnus Christensson,
 *                    Stefan Cronert,
 *                    Stefan Hellkvist,
 *                    Magnus Ljung.
 *
 * Barracuda comes with ABSOLUTELY NO WARRANTY. For detailes see COPYING.
 */

#include <qpainter.h>
#include <qcursor.h>
#include <stdio.h>
#include "MultiLabel.h"

/* MultiLabel
 *
 * Constructor for the MultiLabel object.
 * Constructs a MultiLabel with no columns.
 * Sets the autoresize to FALSE.
 * Input: 
 *    parent      Pointer to the parent widget
 *    name        Name of this widget
 *    Wflags      Flags passed to the QFrame constructor
 */

MultiLabel::MultiLabel (
	QWidget *parent,
	const char *name, 
	WFlags f)
    : QFrame (parent, name, f)
{
    items.setAutoDelete (TRUE); // Delete the columns as they are removed
    autoresize = FALSE;
    scrollOffset = 0;
    lastPressedItem = -1;
    lastPressedSep = -1;
    lastStartX = -1;
    drag = FALSE;
    doLast = FALSE;
}

/* ~MultiLabel
 *
 * Destructor for the MultiLabel class.
 * Will delete all the columns and their contents.
 */

MultiLabel::~MultiLabel()
{
}


/* insertColumn
 *
 * Inserts a column in the multilabel. If the index is -1
 * the column is appended as the last column. Otherwise
 * the column is inserted at the given index, 0 is the lowest
 * index. If the index is higher than the number of columns, the
 * column is appended.
 *
 * The new column will be visible, left-aligned,vertically 
 * centered, has an automatic margin and with manual resizing.
 *
 * The text may be specified.
 */

void MultiLabel::insertColumn (int index)
{
    insertColumn ("", index);
}

void MultiLabel::insertColumn (const char *text, int index)
{
    MultiLabelItem *mi;
    QPainter p;

    mi = new MultiLabelItem ();
    mi->setLabel(text);
    p.begin (this);
    p.setFont (mi->font());
    mi->setWidth(p.fontMetrics().width(QString(mi->label()) + "M"));
    p.end();

    insertItem (mi, index);
}

/* deleteColumn
 *
 * Deletes the column with the given index and redraws the 
 * multilabel. 
 * Input:
 *    index       Which column to delete
 */

void MultiLabel::deleteColumn (int index)
{
    items.remove (index);
    updateLabels ();
}

/* setColumnLabel
 *
 * Sets the label on a column.
 * Input:
 *    text     The text for the column
 *    index    Which column to change
 */

void MultiLabel::setColumnLabel (const char *text, int index)
{
    MultiLabelItem *mi;

    if ((mi = items.at (index)) == NULL)
	return;

    if (mi->label() == text)
	return;

    mi->setLabel(text);

    if (autoresize)
    {
	adjustWidthAndSize ();
	adjustSize ();
    }
    else
	updateLabels ();
}


/* setColumnWidth
 *
 * Sets the width of a column.
 * Input:
 *    width    The new width of the column
 *    index    Which column to change
 */

void MultiLabel::setColumnWidth (int width, int index)
{
    MultiLabelItem *mi;
    int tot = 0;
    QListIterator<MultiLabelItem> it (items);

    if ((mi = items.at (index)) == NULL)
	return;

    mi->setWidth(width);

    if (autoresize)
    {
	adjustWidthAndSize ();
	adjustSize ();
    }

    // Check if the total width can fit in the shown width
    for (; it.current(); ++it)
	if (it.current()->visible())
	    tot += it.current()->width();

    // If all visible columns can fit in the widget - do it.
    if (tot <= QWidget::width())
	setScrollOffset (0);

    emit widthChanged (mi->width(), index);
}

/* setColumnAlignment
 *
 * Sets the alignment of a column.
 * Input:
 *    alignment   The new alignment flags
 *    index       Which column to change
 */

void MultiLabel::setColumnAlignment (int alignment, int index)
{
    MultiLabelItem *mi;

    if ((mi = items.at (index)) == NULL)
	return;

    mi->setAlignment(alignment);
    updateLabels ();
}

/* setColumnMargin
 *
 * Sets the margin of a column.
 * If the margin is negative the margin will be calculated
 * automatically.
 * Input:
 *    margin      The new margin
 *    index       Which column to change
 */

void MultiLabel::setColumnMargin (int margin, int index)
{
    MultiLabelItem *mi;

    if ((mi = items.at (index)) == NULL)
	return;

    mi->setMargin(margin);
}


/* setColumnVisible
 *
 * Determines if the given column should be visible.
 * Redraws the multilabel is there is a change.
 * Input:
 *    visible      TRUE/FALSE
 *    index        Index of column
 */

void MultiLabel::setColumnVisible (bool visible, int index)
{
    MultiLabelItem *mi;
    
    if ((mi = items.at (index)) == NULL)
	return;

    if (mi->visible() != visible)
    {
	mi->setVisible(visible);
	updateLabels ();
    }
}


/* setColumnFont
 *
 * Sets the font for a givent column.
 * Input:
 *    f            The QFont
 *    index        Index of column
 */

void MultiLabel::setColumnFont (const QFont &f, int index)
{
    MultiLabelItem *mi;
    
    if ((mi = items.at (index)) == NULL)
	return;

    mi->setFont(f);
    if (autoresize)
    {
	adjustWidthAndSize ();
	adjustSize ();
    }
    else
	updateLabels ();
}


/* setWidth, setAlignment, setMargin, setFont
 *
 * Like their column counterparts, only that they work on
 * every column.
 */

void MultiLabel::setWidth (int width)
{
    QListIterator<MultiLabelItem> it(items);
    
    for ( ; it.current(); ++it)
	it.current()->setWidth(width);
    
    if (autoresize)
    {
	adjustWidthAndSize ();
	adjustSize ();
    }
}

void MultiLabel::setAlignment (int alignment)
{
    QListIterator<MultiLabelItem> it(items);
    
    for ( ; it.current(); ++it)
	it.current()->setAlignment(alignment);
    
    updateLabels ();
}

void MultiLabel::setMargin (int margin)
{
    QListIterator<MultiLabelItem> it(items);
    
    for ( ; it.current(); ++it)
	it.current()->setMargin(margin);
}


void MultiLabel::setAllFonts (QFont f)
{
    QListIterator<MultiLabelItem> it(items);
    
    for ( ; it.current(); ++it)
	it.current()->setFont(f);

    if (autoresize)
    {
	adjustWidthAndSize();
	adjustSize();
    }
    else
	updateLabels();
}

void MultiLabel::setDoDrag (bool enable)
{
    drag = enable;
}

/* setAutoResize
 *
 * Toggles the autoresize.
 */

void MultiLabel::setAutoResize (bool enable)
{
    if (autoresize != enable)
    {
	autoresize = enable;
	if (autoresize)
	{
	    adjustWidthAndSize ();
	    adjustSize ();
	}
    }
}

/* columnLabel, columnWidth, columnMargin, columnAlignment, 
 * columnVisible, columnFont.
 *
 * Returns the status of that property of the given column.
 * If the column doesn't exist it returns NULL or -1 as appropriate.
 * NOTE: columnVisible returns FALSE if the column didn't exist,
 *       columnFont returns QFont::defaultFont() if no column.
 */

const char *MultiLabel::columnLabel (int index)
{
    MultiLabelItem *mi;
    
    if ((mi = items.at (index)) == NULL)
	return NULL;

    return mi->label();
}

int MultiLabel::columnWidth (int index)
{
    MultiLabelItem *mi;
    
    if ((mi = items.at (index)) == NULL)
	return -1;

    return mi->width();
}

int MultiLabel::columnAlignment (int index)
{
    MultiLabelItem *mi;
    
    if ((mi = items.at (index)) == NULL)
	return -1;

    return mi->alignment();
}

int MultiLabel::columnMargin (int index)
{
    MultiLabelItem *mi;
    
    if ((mi = items.at (index)) == NULL)
	return -1;

    return mi->margin();
}

bool MultiLabel::doDrag ()
{
    return drag;
}

bool MultiLabel::columnVisible (int index)
{
    MultiLabelItem *mi;
    
    if ((mi = items.at (index)) == NULL)
	return FALSE;

    return mi->visible();
}


QFont MultiLabel::columnFont (int index)
{
    MultiLabelItem *mi;
    
    if ((mi = items.at (index)) == NULL)
	return QFont::defaultFont();

    return mi->font();
}

/* scroll
 *
 * Scroll the contents of the frame in the given direction by
 * the given amount.
 * Leftscrolling can only be done until the whole of the leftmost
 * column is shown. 
 * Rightscrolling has no restrictions.
 */

void MultiLabel::scroll (int pixels)
{
    if (scrollOffset + pixels < 0)
	scrollOffset = 0;
    else
	scrollOffset += pixels;

    updateLabels();
}

void MultiLabel::setScrollOffset (int pixels)
{
    // Set scroll offset
    if (scrollOffset + pixels < 0)
	scrollOffset = 0;
    else
	scrollOffset = pixels;
    
    // Announce that the scrolling has changed
    emit scrollChanged (scrollOffset);

    // Repaint labels
    updateLabels();
}
    



/* adjustWidthAndSize
 *
 * Adjusts the widths for all columns so that the 
 * labels fit in nicely.
 */

void MultiLabel::adjustWidthAndSize ()
{
    QListIterator<MultiLabelItem> it(items);
    QPainter p;

    p.begin (this);
    for ( ; it.current(); ++it)
    {
	MultiLabelItem *m = it.current();
	p.setFont (m->font());
	m->setWidth (p.fontMetrics().width (QString(m->label()) + "M"));
    }
    resize (width(), p.fontMetrics().height ());
    p.end ();
}


/* insertItem
 *
 * Inserts a MultiLabelItem object into the list
 */

void MultiLabel::insertItem (MultiLabelItem *m, int index)
{
    if (index < 0 || (items.count() < (unsigned) index))
    {
	items.append (m);
	emit widthChanged (m->width(), items.count()-1);
    }
    else
    {
	items.insert (index, m);
	emit widthChanged (m->width(), index);
    }
}


/* drawContents
 *
 * Draws the contents of the label using a given painter.
 */

void MultiLabel::drawContents (QPainter *p)
{
    QListIterator<MultiLabelItem> it(items);
    QRect cr = contentsRect();
    int xstart = cr.left() - scrollOffset;
    int counter = 0;

    p->setClipRect (cr);
    for (; it.current(); ++it)
	if ((it.current())->visible())
	{
	    MultiLabelItem *c = it.current();
	    
	    QRect r(xstart, cr.top(), c->width(), cr.height());
	    c->drawContents (r, colorGroup().text(), p);
	    xstart += c->width();
	    
	    // Draw the separator 
	    if (((uint) counter < items.count()-1) || doLast)
	    {
		p->setPen (colorGroup().light ());
		p->drawLine (xstart-1, cr.top(), xstart-1, cr.bottom());
		p->setPen (colorGroup().mid ());
		p->drawLine (xstart, cr.top(), xstart, cr.bottom());
	    }
	    xstart++;
	    counter++;
	}
}

void MultiLabel::updateLabels ()
{
    QPainter paint;
    paint.begin (this);
    paint.eraseRect (contentsRect());
    drawContents (&paint);
    paint.end ();
}

/* itemContains
 *
 * Returns the index of the item that contains the given x coord.
 */

int MultiLabel::itemContains (int x)
{
    QListIterator<MultiLabelItem> it(items);
    int startx = (contentsRect().left()) - scrollOffset;
    int counter = 0;
    MultiLabelItem *c;

    if (x < 0)
	x = 0;
    if (x > width())
	x = width();

    for (; it.current(); ++it, counter++)
	if ((it.current())->visible())
	{
	    c = it.current();
	    if (x >= startx && x < (startx + c->width()))
		return counter;
	    
	    startx += c->width () + 1;
	}

    return -1;
}


/* closestLeft
 *
 * Args: An x coordinate somewhere in the widget.
 * => The separator closest to its left.
 */

int MultiLabel::closestLeft (int x)
{
    QListIterator<MultiLabelItem> it(items);
    int startx = (contentsRect().left()) - scrollOffset;
    MultiLabelItem *c;

    for (; it.current(); ++it)
	if ((it.current())->visible())
	{
	    c = it.current();
	    if (x >= startx && x < startx + c->width ())
		return startx;
	    startx += c->width () + 1;
	}

    if (x < 0)
	return 1;
    else
	return startx - c->width() - 1;
}


/* mousePressEvent 
 *
 * Invoked when a mouse button is pressed in the widget.
 * Sets one of lastPressed* and resets the other one.
 */

void MultiLabel::mousePressEvent (QMouseEvent *e)
{
    QListIterator<MultiLabelItem> it(items);
    int startx = (contentsRect().left()) - scrollOffset;
    int counter = 0;
    
    for (; it.current(); ++it, counter++)
	if ((it.current())->visible())
	{
	    MultiLabelItem *c = it.current();

	    if ((e->pos().x() > startx + SEP_WIDTH) && 
		(e->pos().x() < (startx + c->width() - SEP_WIDTH)))
	    {
		lastPressedItem = counter;
		lastPressedSep = -1;
		lastStartX = startx;
		lastPressedX = e->pos().x();
		lastPressedY = e->pos().y();
		doErase = FALSE;
		emit pressed (counter, e->button());
		return;
	    }

	    if ((e->pos().x() > (startx + c->width() - SEP_WIDTH)) &&
		(e->pos().x() < (startx + c->width() + SEP_WIDTH)))
	    {
		lastPressedItem = -1;
		lastPressedSep = counter;
		lastStartX = startx;
		return;
	    }
	    startx += c->width () + 1;
	}
}


/* mouseReleaseEvent 
 *
 * Invoked when a mouse button is release in the widget.
 * Resets lastPressed* and emits appropriate signals.
 */

void MultiLabel::mouseReleaseEvent (QMouseEvent *e)
{
    MultiLabelItem *c;
    
    // Restore the cursor
    setCursor (arrowCursor);
    
    /* Was this the end of a width change? */
    if (lastPressedSep != -1)
    {
	c = items.at (lastPressedSep);
	setColumnWidth (c->width(), lastPressedSep);
	lastPressedItem = -1;
	lastPressedSep = -1;
	lastStartX = -1;
    }

    /* Was this a release on a label? */
    else if (lastPressedItem != -1 && !doErase)
    {
	c = items.at (lastPressedItem);
	/* Was the release over the last pressed item? */
	if (e->pos().x() > lastStartX &&
	    e->pos().x() < lastStartX + c->width())
	    emit clicked (lastPressedItem, e->button());
	emit released (lastPressedItem);
	lastPressedItem = -1;
	lastPressedSep = -1;
	lastStartX = -1;
    }
    else if (lastPressedItem != -1 && doErase)
    {
	MultiLabelItem *mi;
	int after = itemContains (oldLeft);
	if (after == -1 && oldLeft > 1)
	    after = items.count ();
	else if (after == -1)
	    return;
	items.setAutoDelete (FALSE);
	mi = items.at (lastPressedItem);
	items.remove (lastPressedItem);
	if (after >= (int) items.count ())
	    items.append (mi);
	else
	    items.insert (after, mi);
	items.setAutoDelete (TRUE);
	repaint ();
	emit indexChanged (lastPressedItem, after);
	lastPressedItem = -1;
	lastPressedSep = -1;
	lastStartX = -1;
    }

}
	

/* mouseMoveEvent 
 *
 * Invoked when a mouse button is pressed and the mouse is moved.
 */

void MultiLabel::mouseMoveEvent (QMouseEvent *e)
{
    MultiLabelItem *c;
    int w;
    int newLeft;
    QPainter p;
    

    // Are we moving a separator
    if (lastPressedSep != -1)
    {
	setCursor (sizeHorCursor);
	c = items.at (lastPressedSep);
	if (lastStartX != -1 && c != NULL)
	{
	    w = e->pos().x() - lastStartX;
	    if (w < SEP_WIDTH)
		c->setWidth (SEP_WIDTH);
	    else
		c->setWidth (w);
	    updateLabels();
	    if (drag)
		emit widthChanged (c->width(), lastPressedSep);
	}
    }
    // Are we moving a column
    else if (lastPressedItem != -1)
    {
	setCursor (sizeAllCursor);
	newLeft = closestLeft (e->pos().x());
	p.begin (this);
	p.setRasterOp (NotXorROP);
	c = items.at (lastPressedItem);
	if (lastStartX != -1 && c != NULL)
	{
	    if (doErase)
	    {
		p.drawLine (oldLeft, 0, oldLeft, height ());
		p.drawLine (oldLeft, 0, oldLeft + SEP_WIDTH, height() / 2);
		p.drawLine (oldLeft, height(), 
			    oldLeft + SEP_WIDTH, height() / 2);
	    }
	    else
		doErase = TRUE;
	    oldLeft = newLeft;
	    p.drawLine (oldLeft, 0, oldLeft, height ());
	    p.drawLine (oldLeft, 0, oldLeft + SEP_WIDTH, height() / 2);
	    p.drawLine (oldLeft, height(), 
			oldLeft + SEP_WIDTH, height() / 2);
	}
	p.end ();
    }

}


/* mouseDoubleClickEvent
 *
 * Invoked when the widget is double clicked somewhere
 */

void MultiLabel::mouseDoubleClickEvent (QMouseEvent *e)
{
    QListIterator<MultiLabelItem> it(items);
    int startx = (contentsRect().left()) - scrollOffset;
    int counter = 0;
    
    for (; it.current(); ++it, counter++)
	if ((it.current())->visible())
	{
	    MultiLabelItem *c = it.current();

	    if ((e->pos().x() > startx + SEP_WIDTH) && 
		(e->pos().x() < (startx + c->width() - SEP_WIDTH)))
	    {
		emit dblClicked (counter);
		return;
	    }
	    else if ((e->pos().x() > (startx + c->width() - SEP_WIDTH)) &&
		     (e->pos().x() < (startx + c->width() + SEP_WIDTH)))
	    {
		emit sepDblClicked (counter);
		return;
	    }
	    startx += c->width () + 1;
	}
}
  
