/* shortcut.cc
 *
 * Implements the shortcut class, which is responsible for mapping 
 * shortcut keys to functions. The shortcuts may be several keypresses.
 *
 * Barracuda 1.0.
 * Copyright (C) 1997 Magnus Christensson,
 *                    Stefan Cronert,
 *                    Stefan Hellkvist,
 *                    Magnus Ljung.
 *
 * Barracuda comes with ABSOLUTELY NO WARRANTY. For detailes see COPYING.
 */

#include <qapp.h>
#include <qwidget.h>
#include <qsignal.h>
#include "Shortcut.h"


/* ShortcutItem
 *
 * Constructor. Copies the keys, n and id.
 */

ShortcutItem::ShortcutItem (int keys[], int n, int id)
{
    this->n = n;
    this->id = id;
    this->keys = new int[n];

    int i;
    for (i = 0; i < n; i++)
	this->keys[i] = keys[i];

    this->enabled = TRUE;
    this->signal = NULL;
}


/* Shortcut
 *
 * Constructor. Creates a Shortcut object with a parent widget and
 * a name. It doesn't make any sense to make a shortcut object
 * without a parent, since that is where it gets its keyevents from.
 */

Shortcut::Shortcut (QWidget *parent, const char *name)
    : QObject (parent, name)
{
    int i;

    for (i = 0; i < MAXKEYS; i++)
	curKeys[i] = 0;
    nextPos = 0;

    items = new QList<ShortcutItem>;
    items->setAutoDelete (TRUE);

    scBuffer = QString ("");
    cancelKey = CTRL + Key_G;	// Make C-g the default cancel key

    timer = new QTimer (this);  // Create timer
    isTiming = FALSE;

    isRecording = FALSE;

    QObject::connect (timer, SIGNAL (timeout()),
		      this, SLOT (clearBuffer ()));

    if (parent && parent->isWidgetType())
    {
	tlw = parent->topLevelWidget ();
	tlw->installEventFilter (this);
	connect (tlw, SIGNAL (destroyed()), SLOT (tlwDestroyed ()));
    }
    else
    {
	tlw = NULL;
	fprintf (stderr, 
		 "Warning: A Shortcut object should have a parent\n");
    }
}

/* ~Shortcut
 *
 * Destructor.
 */

Shortcut::~Shortcut ()
{
    if (tlw != NULL)
	tlw->removeEventFilter (this);
    
    delete timer;
    delete items;
}


/* setCancelKey
 *
 * Sets the cancel key.
 */

void Shortcut::setCancelKey (int key)
{
    cancelKey = key;
}

/* insertItem
 *
 * Inserts an item with the shortcut given in the keys vector.
 * There should be n entries in keys and keys[0] is the first key and 
 * keys[1] the second and so forth. To identify this shortcut id is used.
 * 
 * The override variable is used to define what happens when the user 
 * tries to define (e.g) C-c when C-cC-c already is defined. If override
 * is TRUE the last definition will overtake the previous one and if it
 * is FALSE insertItem will return FALSE if the above example is tried.
 *
 * Note that if override is TRUE and C-c is defined as a key, then all
 * sequences beginning with C-c will be lost.
 *
 * To define a dummy item (to be changed later) use n = 0.
 */

bool Shortcut::insertItem (int keys[], 
			   int n, 
			   int id, 
			   bool override)
{
    ShortcutItem *s = new ShortcutItem (keys, n, id);
    CHECK_PTR (s);

    QListIterator<ShortcutItem> it(*items);
    int cmp;

    for (; it.current(); ++it)
    {
	cmp = compareItems (it.current(), s);
	if (cmp != SC_OK && override)
	    items->remove (it.current());
	else if (cmp != SC_OK && !override)
	    return FALSE;
    }
    items->append (s);
    return TRUE;
}


/* changeItem
 *
 * Changes the key sequence for a given id.
 * The override variable works as in insertItem.
 * => TRUE if ok FALSE otherwise.
 */

bool Shortcut::changeItem (int keys[], int n, int id, bool override)
{
    ShortcutItem *i = findItem (id);
    QListIterator<ShortcutItem> it(*items);
    ShortcutItem  s (keys, n, id);
    int cmp;
    int k;
    
    if (i == NULL)
	return FALSE;

    if (identical (&s, i))
	return TRUE;

    for (; it.current(); ++it)
    {
	cmp = compareItems (it.current(), &s);
	if (cmp != SC_OK && override)
	    items->remove (it.current());
	else if (cmp != SC_OK && !override)
	    return FALSE;
    }

    delete i->keys;
    i->keys = new int[n];
    CHECK_PTR (i->keys);

    i->n = n;
    for (k = 0; k < n; k++)
	i->keys[k] = keys[k];

    return TRUE;
}


/* clear
 *
 * Clears all the key sequence entries.
 */   

void Shortcut::clear ()
{
    items->clear();
}

/* startRecording
 *
 * Starts the recording of keystrokes - this continues until 
 * stopRecording is called.
 */

void Shortcut::startRecording ()
{
    isRecording = TRUE;
}

/* stopRecording
 *
 * Stops recording.
 * => The number of keystrokes recorded, 
 *    keys, a vector with the keycodes, should be allocated by the user.
 */

int Shortcut::stopRecording (int *keys, int maxkeys)
{
    int i, res;

    isRecording = FALSE;
    res = nextPos;
    for (i = 0; i < nextPos; i++)
    {
	if (i < maxkeys)
	    keys[i] = curKeys[i];
	curKeys[i] = 0;
    }
    nextPos = 0;
    scBuffer = QString ("");
    emit bufferChanged ((const char *)scBuffer);
    
    return res;
}

    
/* isItemEnabled
 *
 * Returns the current state for an item.
 */

bool Shortcut::isItemEnabled (int id)
{
    ShortcutItem *i = findItem (id);

    if (i != NULL)
	return i->enabled;

    return FALSE;
}

/* setItemEnabled
 *
 * Sets the current state for an item.
 */

void Shortcut::setItemEnabled (int id, bool enable)
{
    ShortcutItem *i = findItem (id);

    if (i != NULL)
	i->enabled = enable;
}


/* connectItem
 *
 * Connects an item with a slot. The same item may be connected
 * to several slots.
 */

bool Shortcut::connectItem (int id, 
			    const QObject *receiver, 
			    const char *member)
{
    ShortcutItem *i = findItem (id);
    
    if (i != NULL)
    {
	if (i->signal == NULL)
	{
	    i->signal = new QSignal;
	    CHECK_PTR (i->signal);
	}
	return i->signal->connect (receiver, member);
    }
    
    return FALSE;
}


/* disconnectItem
 *
 * Disconnects an item from a slot. The receiver and the member
 * must be specified since the same signal can be used with many slots.
 */

bool Shortcut::disconnectItem (int id,
			       const QObject *receiver, 
			       const char *member)
{
    ShortcutItem *i = findItem (id);
    
    if (i && i->signal)
	return i->signal->disconnect (receiver, member);
    return FALSE;
}


/* print
 *
 * Prints the shortcuts to a file
 */

void Shortcut::print (FILE *fp)
{
    QListIterator<ShortcutItem> it(*items);

    for (; it.current(); ++it)
    {
	fprintf (fp, "id: %d\n", it.current()->id);
	fprintf (fp, "n: %d\n", it.current()->n);
	for (int j = 0; j < it.current()->n; j++)
	    fprintf (fp, "%s ", (const char *)Shortcut::keyToString (it.current()->keys[j]));
	fprintf (fp, "\n");
    }
}

/* eventFilter
 *
 * Processes keypress events intended for the top level widget.
 */

bool Shortcut::eventFilter (QObject *, QEvent *e)
{
    if (e->type() == Event_Accel)
    {
	if (isTiming)
	    timer->stop ();
	QKeyEvent *k = (QKeyEvent *)e;
	int key = k->key();
	
	if (k->state() & ShiftButton)
	    key |= SHIFT;
	if (k->state() & ControlButton)
	    key |= CTRL;
	if (k->state() & AltButton)
	    key |= ALT;
	
	if (isRecording)
	{
	    addKey (key);
	    k->accept ();
	    return TRUE;
	}
	else
	{
	    if (key == cancelKey)
	    {
		cancelBuffer();
		k->accept ();
		return TRUE;
	    }
	    else
	    {
		addKey (key);
		if (checkSequence () != SC_UNDEFINED)
		{
		    k->accept ();
		    return TRUE;
		}
		return FALSE;
	    }
	}
    }
    return FALSE;
}

/* tlwDestroyed
 *
 * This slot is called when the top level widget that owns the shortcut
 * is destroyed
 */

void Shortcut::tlwDestroyed ()
{
    tlw = NULL;
}

/* clearBuffer
 *
 * Clears the buffer of previous keys.
 */    

void Shortcut::clearBuffer ()
{
    if (scBuffer != "")
    {
	int i;
	for (i = 0; i < MAXKEYS; i++)
	    curKeys[i] = 0;
	nextPos = 0;
	scBuffer = QString ("");
	emit bufferChanged ((const char *)scBuffer);
    }
}


/* addKey
 *
 * Adds a key to the buffer with the previous shortcut keys.
 */

void Shortcut::addKey (int key)
{
    QString keyString = Shortcut::keyToString (key);

    if (keyString == "")
	return;

    curKeys[nextPos++] = key;

    if (scBuffer == "")
    {
	scBuffer = keyString;
	emit bufferChanged ((const char *)scBuffer);
    }
    else
    {
	scBuffer.append (" " + keyString);
	emit bufferChanged ((const char *)scBuffer);
    }
}

/* cancelBuffer
 *
 * Empties the buffer and prints the cancel message instead.
 * After CANCEL_TIME_OUT milliseconds the buffer is cleared if
 * no keys have been pressed before that.
 */

void Shortcut::cancelBuffer ()
{
    int i;
    for (i = 0; i < MAXKEYS; i++)
	curKeys[i] = 0;
    scBuffer = QString ("Cancel!");
    emit bufferChanged ((const char *)scBuffer);
    timer->start (CANCEL_TIME_OUT, TRUE);
    isTiming = TRUE;
}
    
/* keyToString
 *
 * Returns the string corresponding to the key key. The form of 
 * the string is C-c for CTRL-c and M-x for ALT-x, for example.
 */

QString Shortcut::keyToString (int key)
{
    QString res;
    if ((key & ALT) == ALT)
        res += "M-";
    if ((key & CTRL) == CTRL)
	res += "C-";
    if ((key & SHIFT) == SHIFT)
      res += "S-";

    key &= ~(SHIFT | CTRL | ALT);

    QString p;
    if (key >= Key_F1 && key <= Key_F24)
        p.sprintf ("F%d", key - Key_F1 + 1);
    else if (key >= Key_Space && key <= Key_AsciiTilde)
        p.sprintf ("%c", tolower(key));
    else 
    {
        switch (key) 
	{
	case Key_Escape:
	    p = "ESC";
	    break;
	case Key_Tab:
	    p = "Tab";
	    break;
	case Key_Backtab:
	    p = "Backtab";
	    break;
	case Key_Backspace:
	    p = "Backspace";
	    break;
	case Key_Return:
	    p = "Return";
	    break;
	case Key_Enter:
	    p = "Enter";
	    break;
	case Key_Insert:
	    p = "Ins";
	    break;
	case Key_Delete:
	    p = "Del";
	    break;
	case Key_Pause:
	    p = "Pause";
	    break;
	case Key_Print:
	    p = "Print";
	    break;
	case Key_SysReq:
	    p = "SysReq";
	    break;
	case Key_Home:
	    p = "Home";
	    break;
	case Key_End:
	    p = "End";
	    break;
	case Key_Left:
	    p = "Left";
	    break;
	case Key_Up:
	    p = "Up";
	    break;
	case Key_Right:
	    p = "Right";
	    break;
	case Key_Down:
	    p = "Down";
	    break;
	case Key_Prior:
	    p = "PgUp";
	    break;
	case Key_Next:
	    p = "PgDown";
	    break;
	case Key_CapsLock:
	    p = "CapsLock";
	    break;
	case Key_NumLock:
	    p = "NumLock";
	    break;
	case Key_ScrollLock:
	    p = "ScrollLock";
	    break;
	default:
//	    p.sprintf( "<%d?>", key );
	    return "";
//	    break;
        }
    }
    res += p;
    return res;
}

/* stringToKey
 *
 * Args: string with key
 * => corresponding key, -1 on error. pointer pointing after the SC_MATCHed string
 */
int stringToKey (char* string, char** newString)
{
  int key = 0;
  int i;

  /* Semi-Feature: The key strings are case sensitive */

  static struct {
    int keysym;
    char* str;
  } keys[] = {
    {Key_Escape, "ESC"},
    {Key_Tab, "Tab"},
    {Key_Backtab, "Backtab"},
    {Key_Backspace, "Backspace"},
    {Key_Return, "Return"},
    {Key_Enter, "Enter"},
    {Key_Insert, "Ins"},
    {Key_Delete, "Del"},
    {Key_Pause, "Pause"},
    {Key_Print, "Print"},
    {Key_SysReq, "SysReq"},
    {Key_Home, "Home"},
    {Key_End, "End"},
    {Key_Left, "Left"},
    {Key_Up, "Up"},
    {Key_Right, "Right"},
    {Key_Down, "Down"},
    {Key_Prior, "PgUp"},
    {Key_Next, "PgDown"},
    {Key_CapsLock, "CapsLock"},
    {Key_NumLock, "NumLock"},
    {Key_ScrollLock, "ScrollLock"},
    {0, (char*) 0}};

  /* Check for modifiers */
  if (!strncmp ("M-", string, 2))
    {
      string += 2;
      key |= ALT;
    }
  if (!strncmp ("C-", string, 2))
    {
      string += 2;
      key |= CTRL;
    }
  if (!strncmp ("S-", string, 2))
    {
      string += 2;
      key |= SHIFT;
    }

  /* Check for special keys */
  for (i = 0; keys[i].str; i++)
    if (!strncmp (keys[i].str, string, strlen (keys[i].str)))
      {
	key |= keys[i].keysym;
	string += strlen (keys[i].str);
	*newString = string;
	return key;
      }

  /* Check for F keys */
  if (string[0] == 'F')
    {
      int fkeynum;
      string++;
      for (fkeynum = 0; isdigit (string[0]); string++)
	fkeynum = fkeynum*10 + string[0] - '0';

      if (fkeynum < 1 || fkeynum > 24)
	return -1;
      key |= Key_F1 + fkeynum - 1;
      *newString = string;
      return key;
    }

  /* Check for normal keys */
  if (string[0] >= ' ' && string[0] <= '~')
    {
      key |= toupper (string[0]);
      string++;
      *newString = string;
    }
  else
    return -1;

  return key;
}

/* compareItems
 *
 * Compares two shortcut items. If the first one starts with the second
 * one it returns UNDERRIDE. If the first one starts with a part of the
 * second one it returns OVERRIDE. If they're equal one of the two are
 * returned. Otherwise SC_OK is returned.
 */

int Shortcut::compareItems (ShortcutItem *i1, ShortcutItem *i2)
{
    int i;

    if (i2->n == 0)
	return SC_OK;

    for (i = 0; i < i2->n; i++)
    {
	if (i1->keys[i] != i2->keys[i])
	    return SC_OK;
	if (i1->keys[i] == i2->keys[i] && i1->n == i)
	    return OVERRIDE;
    }
    
    return UNDERRIDE;
}


/* findItem
 *
 * Returns the first item found with the given id.
 * Returns NULL on failure.
 */

ShortcutItem *Shortcut::findItem (int id)
{
    QListIterator<ShortcutItem> it(*items);

    for (; it.current(); ++it)
	if (it.current()->id == id)
	    return it.current();

    return NULL;
}
	

/* compareItem
 *
 * Compares the item's key sequence with the current one.
 * Returns SC_MATCH if they SC_MATCH, SC_PARTIAL if it is the beginning
 * of this shortcut and SC_UNDEFINED otherwise.
 */

int Shortcut::compareItem (ShortcutItem *i)
{
    int k;

    if ((nextPos > i->n) || (i->n == 0))
	return SC_UNDEFINED;
    
    for (k = 0; k < nextPos; k++)
	if (i->keys[k] != curKeys[k])
	    return SC_UNDEFINED;

    if (i->n == nextPos)
	return SC_MATCH;

    return SC_PARTIAL;
}


/* checkSequence
 *
 * Checks if the current sequence should give any signals.
 */

int Shortcut::checkSequence ()
{
    QListIterator<ShortcutItem> it(*items);
    int res;
    bool undef = TRUE;
    int i;
    
    for (; it.current(); ++it)
    {
	res = compareItem (it.current());
	if (res == SC_MATCH)
	{
	    undef = FALSE;
	    clearBuffer();
	    if (it.current()->enabled)
	      {
		if (it.current()->signal)
		  it.current()->signal->activate();
		else
		  emit activated (it.current()->id);
	      }
	}
	else if (res == SC_PARTIAL)
	    undef = FALSE;
    }

    if (undef == TRUE)
    {
	for (i = 0; i < MAXKEYS; i++)
	    curKeys[i] = 0;
	clearBuffer ();
	return SC_UNDEFINED;
    }
    
    return res;
}
	    
/* identical
 *
 * Args: i1, i2
 * => TRUE if i1 and i2 has the same id's and key sequence.
 */

bool Shortcut::identical (ShortcutItem *i1, ShortcutItem *i2)
{
    int k;

    if (i1->id == i2->id && i1->n == i2->n)
    {
	for (k = 0; k < i1->n; k++)
	    if (i1->keys[k] != i2->keys[k])
		return FALSE;
	return TRUE;
    }

    return FALSE;
}
