/* MBMessageBox.cc
 *
 * The MBMessageBox class implements a dialog which has a message and a 
 * dynamic number of buttons.
 *
 * Barracuda 1.0.
 * Copyright (C) 1997 Magnus Christensson,
 *                    Stefan Cronert,
 *                    Stefan Hellkvist,
 *                    Magnus Ljung.
 *
 * Barracuda comes with ABSOLUTELY NO WARRANTY. For detailes see COPYING.
 */

#include "MBMessageBox.h"

/* MBMessageBox
 *
 * Constructor for the MBMessageBox
 */

MBMessageBox::MBMessageBox (QWidget *parent, const char *name)
    : QDialog (parent, name, TRUE)
{
    label = new QLabel (this, "text");
    CHECK_PTR (label);
    label->setAlignment (AlignCenter);
    QFont font ("Helvetica", 12);
    label->setFont (font);
    num = 0;
}

/* ~MBMessageBox
 *
 * Destructor, destroys the label and all buttons.
 */

MBMessageBox::~MBMessageBox ()
{
    int i;

    for (i = 0; i < numButtons(); i++)
	delete buttons[i];

    delete buttons;
    delete label;
}


/* text
 *
 * => The message text.
 */

const char *MBMessageBox::text ()
{
    return label->text();
}

/* setText
 *
 * Sets the message text.
 */

void MBMessageBox::setText (const char *t)
{
    label->setText (t);
}
 
/* numButtons
 *
 * => The number of buttons in the dialog.
 */   

int MBMessageBox::numButtons ()
{
    return num;
}

/* setNumButtons
 *
 * Sets the number of buttons.
 */

void MBMessageBox::setNumButtons (int n)
{
    int i;
    QFont font ("Helvetica", 12);
    font.setWeight (QFont::Bold);
    QString bl;
    
    if (n < 0 || n > MAXBUTTONS)
    {
	fprintf (stderr, "Warning! A MBMessageBox can only have between "
		 "1 and %d buttons - sorry...\n", MAXBUTTONS);
	if (n < 0)
	    n = 0;
	if (n > MAXBUTTONS)
	    n = MAXBUTTONS;
    }

    // Check if no buttons are required
    if (n == 0)
    {
	for (i = 0; i < num; i++)
	    delete buttons[i];
	delete buttons;
	return;
    }

    // Allocate space for the vector
    if (num < n)
    {
	QPushButton **newButtons = new (QPushButton*)[n];
	CHECK_PTR (newButtons);
	for (i = 0; i < num; i++)
	    newButtons[i] = buttons[i];
	if (num != 0)
	    delete buttons;
	buttons = newButtons;
    }
	
    // Fill in the vector
    if (num == n)
	return;
    else if (num > n)
	for (i = n; i < num; i++)
	    delete buttons[i];
    else if (num < n)
	for (i = num; i < n; i++)
	{
	    bl.sprintf ("Btn%d", i);
	    buttons[i] = new QPushButton (bl, this);
	    CHECK_PTR (buttons[i]);
	    buttons[i]->setFont (font);
	    connectButton (i);
	}

    num = n;
}
    
/* buttonText
 *
 * => The text on a given button.
 */
    
const char *MBMessageBox::buttonText (int button)
{
    if (button < 0 || button >= numButtons())
	return NULL;

    return buttons[button]->text();
}

/* setButtonText
 *
 * Sets the button text on a given button.
 */

void MBMessageBox::setButtonText (const char *text, int button)
{
    if (button < 0 || button >= numButtons ())
	return;

    buttons[button]->setText (text);
    buttons[button]->adjustSize ();
}

/* adjustSize
 *
 * Adjusts the size of the message box to fit the contents.
 */

void MBMessageBox::adjustSize ()
{
    QSize max, cur;
    int i;
    
    if (numButtons() > 0)
    {
	// Find the size of the largest button
	max = buttons[0]->sizeHint();
	for (i = 1; i < numButtons(); i++)
	{
	    cur = buttons[i]->sizeHint ();
	    max = QSize (QMAX (cur.width(), max.width()),
			 QMAX (cur.height(), max.height()));
	}
	
	// Set all button sizes to the largest size
	for (i = 0; i < numButtons(); i++)
	    buttons[i]->resize (max);
    }

    // Adjust the label size
    label->adjustSize ();

    // Calculate how big the dialog needs to be
    QString labelStr = label->text ();
    int nlines = labelStr.contains ('\n');
    QFontMetrics fm = label->fontMetrics ();
    nlines += 2;
    int bw = 0; 
    for (i = 0; i < numButtons (); i++)
	bw += buttons[i]->width() + 5;
    int w = QMAX (bw, label->width ());
    int h = fm.lineSpacing() * nlines;
    if (numButtons() > 0)
	h += buttons[0]->height ();

    resize (w + w/3, h + h/3);
}
    
    
/* resizeEvent
 *
 * Handles reszize events
 */    

void MBMessageBox::resizeEvent (QResizeEvent *)
{
    QSize max, cur;
    int i;
    
    if (numButtons() > 0)
    {
	// Find the size of the largest button
	max = buttons[0]->sizeHint();
	for (i = 1; i < numButtons(); i++)
	{
	    cur = buttons[i]->sizeHint ();
	    max = QSize (QMAX (cur.width(), max.width()),
			 QMAX (cur.height(), max.height()));
	}
	
	// Set all button sizes to the largest size
	for (i = 0; i < numButtons(); i++)
	    buttons[i]->resize (max);
    }

    // Adjust the label size
    label->adjustSize ();

    // Place the label
    int h;
    if (numButtons() > 0)
	h = (height() - buttons[0]->height() - label->height())/3;
    else
	h = (height() - label->height())/3;
    label->move (width()/2 - label->width()/2, h);

    // Place the buttons
    int bw = 0;
    int w;
    int n = numButtons() + 1;
    for (i = 0; i < numButtons(); i++)
	bw += buttons[i]->width();
    w = width() - bw;
    for (i = 0; i < numButtons(); i++)
	buttons[i]->move (i*buttons[i]->width() + (i+1)*w/n, 
			  height() - h - buttons[i]->height());
}


/* query
 *
 * Opens a modal message box and returns an index indicating which button
 * was pressed.
 * After numOfButtons, that number of strings should be supplied
 * that are the button labels.
 */

int MBMessageBox::query (const char *caption,
			 const char *text,
			 int numOfButtons,
			 ...)
{
    int i;
    va_list ap;
    MBMessageBox *mb = new MBMessageBox ();
    CHECK_PTR (mb);

    mb->setCaption (caption);
    mb->setText (text);
    mb->setNumButtons (numOfButtons);
    va_start (ap, numOfButtons);
    for (i = 0; i < numOfButtons; i++)
	mb->setButtonText (va_arg (ap, char *), i);
    va_end (ap);

    int retcode = mb->exec ();
    delete mb;

    return retcode;
}
    




/* Slots for buttons - this is a kludge but I don't
 * know any better way!
 */

void MBMessageBox::connectButton (int i)
{
    switch (i)
    {
    case 0:
	connect (buttons[0], SIGNAL (clicked()), SLOT (buttonPressed1 ()));
	break;
    case 1:
	connect (buttons[1], SIGNAL (clicked()), SLOT (buttonPressed2 ()));
	break;
    case 2:
	connect (buttons[2], SIGNAL (clicked()), SLOT (buttonPressed3 ()));
	break;
    case 3:
	connect (buttons[3], SIGNAL (clicked()), SLOT (buttonPressed4 ()));
	break;
    case 4:
	connect (buttons[4], SIGNAL (clicked()), SLOT (buttonPressed5 ()));
	break;
    default:
	fprintf (stderr, "MBMessageBox: Internal error trying to connect "
		 "to nonexistent slot.\n");
    }
}

void MBMessageBox::buttonPressed1 ()
{
    done (0);
}

void MBMessageBox::buttonPressed2 ()
{
    done (1);
}

void MBMessageBox::buttonPressed3 ()
{
    done (2);
}

void MBMessageBox::buttonPressed4 ()
{
    done (3);
}

void MBMessageBox::buttonPressed5 ()
{
    done (4);
}

